from libcpp.string cimport string

cdef extern from "<string_view>" namespace "std" nogil:
    cdef cppclass string_view:
        string_view()
        string_view(const string_view& other)
        string_view(const char* s, size_t count)
        string_view& operator=(const string_view& view)
        bint empty() const
        size_t size() const
        string_view substr(size_t pos, size_t count) const
        string_view substr(size_t pos) const
        string_view substr() const
        size_t find(const char* s, size_t pos)
        size_t find(const char* s)
        const char& operator[](size_t pos) const
        const char* data() const
        const char& front()
        const char& back()
        string_view remove_prefix(size_t n)
        string_view remove_suffix(size_t n)

    bint operator==(const string_view& lhs, const string_view& rhs)
    bint operator==(const char* lhs, const string_view& rhs)
    bint operator==(const string_view& lhs, const char* rhs)
    bint operator==(const string_view& lhs, const string& rhs)
    bint operator==(const string& lhs, const string_view& rhs)

    bint operator!=(const string_view& lhs, const string_view& rhs)
    bint operator!=(const char* lhs, const string_view& rhs)
    bint operator!=(const string_view& lhs, const char* rhs)
    bint operator!=(const string_view& lhs, const string& rhs)
    bint operator!=(const string& lhs, const string_view& rhs)



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

import typing as t

from fastwarc.warc cimport ArchiveIterator
from resiliparse.process_guard cimport progress


def progress_loop(it, ctx=None) -> t.Iterable[t.Any]:
    """
    progress_loop(it, ctx=None)

    Wraps an iterator into a pass-through iterator that reports progress
    to an active :class:`resiliparse.process_guard.TimeGuard` context guard after each iteration.

    :param it: original iterator
    :type it: t.Iterable[t.Any]
    :param ctx: active guard context (will use last global context from stack if unset)
    :return: wrapped iterator
    :rtype: t.Iterable[t.Any]
    """
    for i in it:
        yield i
        progress(ctx)


def exc_loop(it):
    """
    exc_loop(it)

    Wraps an iterator into another iterator that catches and returns any exceptions raised
    while evaluating the input iterator.

    This is primarily useful for unreliable generators that may throw unpredictably at any time
    for unknown reasons (e.g., generators reading from a network data source). If you do not want to
    wrap the entire loop in a ``try/except`` clause, you can use an :func:`exc_loop` to catch
    any such exceptions and return them.
    Remember that a generator will end after throwing an exception, so if the input iterator is
    a generator, you will have to create a new instance in order to retry or continue.

    :param it: original iterator
    :type it: t.Iterable[t.Any]
    :return: iterator of items and ``None`` or ``None`` and exception instance
    :rtype: t.Iterable[(t.Any or None, BaseException or None)]
    """
    i = iter(it)
    while True:
        try:
            yield next(i), None
        except StopIteration as e:
            raise e
        except BaseException as e:
            yield None, e


def warc_retry(ArchiveIterator archive_iterator, stream_factory, retry_count=3, seek=True):
    """
    warc_retry(archive_iterator, stream_factory, retry_count=3, seek=True)

    Wrap a :class:`fastwarc.warc.ArchiveIterator` to try to continue reading after a stream failure.

    Use if the underlying stream is unreliable, such as when reading from a network data source.
    If an exception other than :exc:`StopIteration` is raised while consuming the iterator, the WARC
    reading process will be retried up to `retry_count` times. When a stream failure occurs,
    ``archive_iterator`` will be reinitialised with a new stream object by calling ``stream_factory``.

    The new stream object returned by ``stream_factory()`` must be seekable. If the stream does not
    support seeking, you can set ``seek=False``. In this case, the stream position in bytes of the last
    successfully read record will be passed as a parameter to ``stream_factory()``. The factory is then
    expected to return a stream that already starts at this exact position (or else reading would
    restart from the beginning resulting in duplicate records). This is primarily useful for streams
    that are not inherently seekable, but have an external facility for starting them at the correct
    position (such as S3 HTTPS streams created from range requests).

    As another option, ``seek`` can also be ``None``, which instructs :func:`warc_retry` to consume the
    stream up to the continuation position. The stream returned by ``stream_factory()`` must start at
    the beginning and will be read normally, but all bytes before the last record will be skipped over
    before continuing to parse the contents. This is the most expensive method of "seeking" on a stream
    and should only be used if the stream is not seekable and there is no other option for starting it
    at the correct offset.

    Exceptions raised inside ``stream_factory()`` will be caught and count towards ``retry_count``.

    :param archive_iterator: input WARC iterator
    :type archive_iterator: fastwarc.warc.ArchiveIterator
    :param stream_factory: callable returning a new stream instance to continue iteration in case of failure
    :type stream_factory: t.Callable[[], t.Any] or t.Callable[[int], t.Any]
    :param retry_count: maximum number of retries before giving up (set to ``None`` or zero for no limit)
    :type retry_count: int
    :param seek: whether to seek to previous position on new stream object (or ``None`` for "stream consumption")
    :type seek: bool or None
    :return: wrapped :class:`~fastwarc.warc.ArchiveIterator`
    :rtype: t.Iterable[fastwarc.warc.WarcRecord]
    """

    cdef size_t retries = 0
    cdef size_t pos = archive_iterator.reader.tell()
    cdef size_t first_pos = pos
    cdef bint skip_next = False
    it = archive_iterator.__iter__()

    while True:
        try:
            if skip_next:
                next(it)
                skip_next = False
            next_rec = next(it)
            yield next_rec
            pos = next_rec.stream_pos
        except StopIteration as e:
            raise e
        except BaseException as e:
            retries += 1
            if retry_count and retries > retry_count:
                raise e

            while True:
                try:
                    if seek is True:
                        stream = stream_factory()
                        stream.seek(pos)
                        break
                    elif seek is False:
                        stream = stream_factory(pos)
                        break
                    elif seek is None:
                        consumed = 0
                        stream = stream_factory()
                        while consumed < pos:
                            n = len(stream.read(min(16384, pos - consumed)))
                            if n == 0:
                                return  # Unexpected EOF
                            consumed += n
                        break

                except BaseException as e:
                    retries += 1
                    if retry_count and retries > retry_count:
                        raise e

            # noinspection PyProtectedMember
            archive_iterator._set_stream(stream)
            it = archive_iterator.__iter__()
            skip_next = pos > first_pos



cdef extern from "<algorithm>" namespace "std" nogil:
    void replace[Iter, T](Iter first, Iter last, const T& old_value, const T& new_value)
    int count[Iter, T](Iter first, Iter last, const T& val)



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

import typing as t

cimport cython
from cython.operator cimport preincrement as preinc, predecrement as predec
from cpython.ref cimport PyObject
from libcpp.set cimport set as unordered_set

from resiliparse_inc.lexbor cimport *
from resiliparse.parse.encoding cimport bytes_to_str, map_encoding_to_html5


cdef inline DOMNode _create_dom_node(HTMLTree tree, lxb_dom_node_t* dom_node):
    if not dom_node:
        return None
    if dom_node.user:
        return <DOMNode>dom_node.user
    cdef DOMNode node = DOMNode.__new__(DOMNode, tree)
    node.node = dom_node
    node.node.user = <PyObject*>node
    return node


cdef inline DOMCollection _create_dom_collection(HTMLTree tree, lxb_dom_collection_t* coll):
    cdef DOMCollection return_coll = DOMCollection.__new__(DOMCollection, tree)
    return_coll.coll = coll
    return return_coll


cdef bint init_css_parser(lxb_css_parser_t** parser) except 0:
    parser[0] = lxb_css_parser_create()
    if lxb_css_parser_init(parser[0], NULL, NULL) != LXB_STATUS_OK:
        raise RuntimeError('Failed to initialize CSS parser.')
    return True


cdef void destroy_css_parser(lxb_css_parser_t* parser) nogil:
    if parser:
        lxb_css_parser_destroy(parser, True)


cdef bint init_css_selectors(lxb_css_parser_t* parser, lxb_css_selectors_t** css_selectors,
                             lxb_selectors_t** selectors) except 0:
    css_selectors[0] = lxb_css_selectors_create()
    if lxb_css_selectors_init(css_selectors[0], 32) != LXB_STATUS_OK:
        raise RuntimeError('Failed to initialize CSS selectors.')

    lxb_css_parser_selectors_set(parser, css_selectors[0])

    selectors[0] = lxb_selectors_create()
    if lxb_selectors_init(selectors[0]) != LXB_STATUS_OK:
        raise RuntimeError('Failed to initialize elemetn selectors.')

    return True


cdef void destroy_css_selectors(lxb_css_selectors_t* css_selectors, lxb_selectors_t* selectors) nogil:
    if selectors:
        lxb_selectors_destroy(selectors, True)
    if css_selectors:
        lxb_css_selectors_destroy(css_selectors, True, True)


cdef inline void _log_serialize_cb(const lxb_char_t *data, size_t len, void *ctx) nogil:
    (<string*>ctx).append(<const char*>data, len)


cdef lxb_css_selector_list_t* parse_css_selectors(lxb_css_parser_t* css_parser, const lxb_char_t* selector,
                                                  size_t selector_len) except NULL:
    cdef lxb_css_selector_list_t* sel_list = lxb_css_selectors_parse(css_parser, selector, selector_len)
    cdef string err
    if css_parser.status != LXB_STATUS_OK:
        lxb_css_log_serialize(css_parser.log, <lexbor_serialize_cb_f>_log_serialize_cb, &err, <const lxb_char_t*>b'', 0)
        raise ValueError(f'CSS parser error: {err.decode().strip()}')
    return sel_list


cdef inline lxb_dom_node_t* next_node(const lxb_dom_node_t* root_node, lxb_dom_node_t* node,
                                      size_t* depth=NULL, bint* end_tag=NULL) nogil:
    """
    DOM tree pre-order traversal primitive.
    
    This is a more flexible, step-wise implementation of ``lxb_dom_node_simple_walk()``, which
    allows a client to react on closing tags.

    :param root_node: root node at which traversal started
    :param node: current node
    :param depth: tracks DOM depth if not ``NULL`` (needs to be initialized and passed back in at each step)
    :param end_tag: if not ``NULL``, tracks whether step was an end tag (needs to be initialized with ``False``
                    and passed back in at each step)
    :returns: next node or ``NULL`` if done
    """
    cdef bint is_end = end_tag and end_tag[0]

    if not is_end and node.first_child:
        if depth:
            preinc(depth[0])
        return node.first_child
    else:
        while not node.next and node != root_node:
            node = node.parent
            if depth:
                predec(depth[0])
            if end_tag:
                end_tag[0] = True
                return node

        if end_tag:
            end_tag[0] = False
        if node == root_node:
            return NULL
        return node.next


cdef inline string get_node_text(lxb_dom_node_t* node) nogil:
    """Get node inner text."""
    cdef lxb_dom_character_data_t* char_data
    if node.type == LXB_DOM_NODE_TYPE_TEXT:
        char_data = <lxb_dom_character_data_t*>node
        return string(<const char*>char_data.data.data, char_data.data.length)

    cdef size_t text_len = 0
    cdef lxb_char_t* text = lxb_dom_node_text_content(node, &text_len)
    if not text or not text_len:
        return string()
    cdef string text_str = string(<const char*>text, text_len)
    lxb_dom_document_destroy_text(node.owner_document, text)
    return text_str


cdef lxb_dom_collection_t* get_elements_by_class_name_impl(lxb_dom_node_t* node, bytes class_name, size_t init_size=5):
    """
    Return a collection of elements matching the given attribute name and value.

    The caller must take ownership of the returned collection.

    :param node: anchor node
    :param class_name: class name as UTF-8 bytes
    :param init_size: initial collection size
    :return: pointer to created DOM collection or ``NULL`` if error occurred
    """
    cdef lxb_dom_collection_t * coll = lxb_dom_collection_make(node.owner_document, init_size)
    if not coll:
        return NULL

    cdef lxb_status_t status = lxb_dom_elements_by_class_name(<lxb_dom_element_t*>node, coll,
                                                              <const lxb_char_t*>class_name, len(class_name))

    if status != LXB_STATUS_OK:
        lxb_dom_collection_destroy(coll, True)
        return NULL

    return coll


cdef struct element_by_id_match_ctx:
    const lxb_char_t* id_val
    size_t id_val_len
    lxb_dom_node_t* result_node

ctypedef element_by_id_match_ctx element_by_id_match_ctx_t


cdef inline lexbor_action_t element_by_id_callback(lxb_dom_node_t* node, void* ctx) nogil:
    if node.type != LXB_DOM_NODE_TYPE_ELEMENT:
        return LEXBOR_ACTION_OK

    cdef element_by_id_match_ctx_t* match_ctx = <element_by_id_match_ctx_t*>ctx
    cdef const lxb_char_t* id_val
    cdef size_t id_val_len = 0

    id_val = lxb_dom_element_id(<lxb_dom_element_t*>node, &id_val_len)
    if not id_val or id_val_len != match_ctx.id_val_len:
        return LEXBOR_ACTION_OK

    if lexbor_str_data_ncmp(id_val, match_ctx.id_val, id_val_len):
        match_ctx.result_node = node
        return LEXBOR_ACTION_STOP

    return LEXBOR_ACTION_OK


cdef lxb_dom_node_t* get_element_by_id_impl(lxb_dom_node_t* node, bytes id_value, bint case_insensitive=False):
    """
    Return a pointer to the first element matching the given ID or NULL.

    :param node: anchor node
    :param id_value: ID value as UTF-8 bytes
    :param case_insensitive: match case-insensitive
    :return: pointer to created DOM collection or ``NULL`` if error occurred
    """

    cdef element_by_id_match_ctx ctx = [<const lxb_char_t*>id_value, len(id_value), NULL]
    lxb_dom_node_simple_walk(node, <lxb_dom_node_simple_walker_f>element_by_id_callback, &ctx)
    return ctx.result_node


cdef lxb_dom_collection_t* get_elements_by_attr_impl(lxb_dom_node_t* node, bytes attr_name, bytes attr_value,
                                                     size_t init_size=5, bint case_insensitive=False):
    """
    Return a collection of elements matching the given attribute name and value.

    The caller must take ownership of the returned collection.

    :param node: anchor node
    :param attr_name: attribute name as UTF-8 bytes
    :param attr_value: attribute value as UTF-8 bytes
    :param init_size: initial collection size
    :param case_insensitive: match case-insensitive
    :return: pointer to created DOM collection or ``NULL`` if error occurred
    """
    cdef lxb_dom_collection_t * coll = lxb_dom_collection_make(node.owner_document, init_size)
    if not coll:
        return NULL

    cdef lxb_status_t status = lxb_dom_elements_by_attr(<lxb_dom_element_t*>node, coll,
                                                        <lxb_char_t*>attr_name, len(attr_name),
                                                        <lxb_char_t*>attr_value, len(attr_value),
                                                        case_insensitive)

    if status != LXB_STATUS_OK:
        lxb_dom_collection_destroy(coll, True)
        return NULL

    return coll


cdef lxb_dom_collection_t* get_elements_by_tag_name_impl(lxb_dom_node_t* node, bytes tag_name):
    """
    Return a collection of elements matching the given tag name.

    The caller must take ownership of the returned collection.
    
    :param node: anchor node
    :param tag_name: tag name as UTF-8 bytes
    """
    cdef lxb_dom_collection_t* coll = lxb_dom_collection_make(node.owner_document, 20)
    if not coll:
        raise RuntimeError('Failed to create DOM collection')

    cdef lxb_status_t status = lxb_dom_elements_by_tag_name(<lxb_dom_element_t*>node,
                                                            coll, <const lxb_char_t*>tag_name, len(tag_name))
    return coll


cdef inline lxb_status_t css_select_callback(lxb_dom_node_t* node, lxb_css_selector_specificity_t* spec,
                                             void* ctx) nogil:
    lxb_dom_collection_append(<lxb_dom_collection_t*>ctx, node)
    return LXB_STATUS_OK


cdef inline lxb_status_t css_select_callback_single(lxb_dom_node_t* node, lxb_css_selector_specificity_t* spec,
                                                    void* ctx) nogil:
    (<lxb_dom_node_t**>ctx)[0] = node
    return LXB_STATUS_STOP


cdef lxb_dom_node_t* query_selector_impl(lxb_dom_node_t* node, HTMLTree tree,
                                         bytes selector) except <lxb_dom_node_t*>-1:
    """
    Return a pointer to the first element matching the given CSS selector.

    :param node: anchor node
    :param tree: owning HTML tree
    :param selector: CSS selector as UTF-8 bytes
    :return: pointer to created DOM collection or ``NULL`` if error occurred
    """
    tree.init_css_parser()

    cdef lxb_css_selector_list_t* sel_list = parse_css_selectors(tree.css_parser,
                                                                 <const lxb_char_t*>selector, len(selector))
    cdef lxb_dom_node_t* result_node = NULL
    if lxb_selectors_find(tree.selectors, node, sel_list,
                          <lxb_selectors_cb_f>css_select_callback_single, &result_node) != LXB_STATUS_OK:
        return NULL

    return result_node


cdef lxb_dom_collection_t* query_selector_all_impl(lxb_dom_node_t* node, HTMLTree tree, bytes selector,
                                                   size_t init_size=32) except <lxb_dom_collection_t*>-1:
    """
    Return a collection of elements matching the given CSS selector.

    The caller must take ownership of the returned collection.

    :param node: anchor node
    :param tree: owning HTML tree
    :param selector: CSS selector as UTF-8 bytes
    :param init_size: initial collection size
    :return: pointer to created DOM collection or ``NULL`` if error occurred
    """
    tree.init_css_parser()

    cdef lxb_css_selector_list_t* sel_list = parse_css_selectors(tree.css_parser,
                                                                 <const lxb_char_t*>selector, len(selector))
    cdef lxb_dom_collection_t* coll = lxb_dom_collection_make(node.owner_document, init_size)
    if lxb_selectors_find(tree.selectors, node, sel_list,
                          <lxb_selectors_cb_f>css_select_callback, coll) != LXB_STATUS_OK:
        return NULL

    return coll


cdef inline lxb_status_t css_match_callback(lxb_dom_node_t* node, lxb_css_selector_specificity_t* spec,
                                            void* ctx) nogil:
    (<bint*>ctx)[0] = True
    return LXB_STATUS_STOP


cdef bint matches_impl(lxb_dom_node_t* node, HTMLTree tree, bytes selector):
    """
    Check whether any element in the DOM subtree matches the given CSS selector.

    :param node: anchor node
    :param tree: owning HTML tree
    :param selector: CSS selector as bytes
    :return: boolean value indicating whether a matching element exists
    """
    tree.init_css_parser()

    cdef lxb_css_selector_list_t* sel_list = parse_css_selectors(tree.css_parser,
                                                                 <const lxb_char_t*>selector, len(selector))
    cdef bint matches = False
    if lxb_selectors_find(tree.selectors, node, sel_list,
                          <lxb_selectors_cb_f>css_match_callback, <void*>&matches) != LXB_STATUS_OK:
        return False

    return matches


cdef inline bint is_whitespace(const char c):
    return c == b' ' or c == b'\t' or c == b'\n' or c == b'\f' or c == b'\r'


cdef class DOMElementClassList:
    """Class name list of an Element DOM node."""

    def __cinit__(self, DOMNode node):
        self.node = node

    cdef list _create_list(self):
        if self.node is None or not self.node.node:
            return []

        cdef size_t class_name_len = 0
        cdef const lxb_char_t* class_name = lxb_dom_element_class(<lxb_dom_element_t*>self.node.node, &class_name_len)
        if not class_name:
            return []

        cdef list class_list = []
        cdef start = 0, end = 0
        cdef size_t i
        for i in range(class_name_len):

            if is_whitespace(class_name[start]):
                start = i + 1
                continue

            if is_whitespace(class_name[i]) or i == class_name_len - 1:
                end = i if i < class_name_len - 1 else i + 1
                if start < end:
                    class_list.append(class_name[start:end].decode())
                    start = i + 1

        return class_list

    cdef inline bytes _class_name_bytes(self):
        cdef size_t class_name_len = 0
        cdef const lxb_char_t* class_name = lxb_dom_element_class(<lxb_dom_element_t*>self.node.node, &class_name_len)
        if not class_name:
            return b''
        return class_name[:class_name_len]

    cpdef void add(self, str class_name):
        """
        add(self, class_name)
        
        Add new class name to Element node if not already present.
        
        :param class_name: new class name
        :type class_name: str
        """
        if self.node is None or not self.node.node:
            return

        cdef list l = self._create_list()
        if class_name in l:
            return

        cdef bytes new_class_name = self._class_name_bytes()
        if not is_whitespace(new_class_name[-1]):
            class_name = ' ' + class_name
        new_class_name = new_class_name + class_name.encode()
        # noinspection PyProtectedMember
        self.node._setattr_impl(b'class', new_class_name)

    cpdef void remove(self, str class_name):
        """
        remove(self, class_name)
        
        Remove a class name from this Element node.
        
        :param class_name: new class name
        :type class_name: str
        """
        if self.node is None or not self.node.node:
            return

        cdef list l = [c for c in self._create_list() if c != class_name]
        # noinspection PyProtectedMember
        self.node._setattr_impl(b'class', b' '.join([c.encode() for c in l]))

    def __contains__(self, item):
        """
        __contains__(self, item):
        """
        return item in self._create_list()

    def __getitem__(self, item):
        """
        __contains__(self, item)
        """
        return self._create_list()[item]

    def __eq__(self, other):
        """
        __eq__(self, other)
        """
        return other == self._create_list()

    def __len__(self):
        """
        __len__(self)
        """
        return len(self._create_list())

    def __iter__(self):
        """
        __iter__(self)
        """
        return iter(self._create_list())

    def __repr__(self):
        """
        __repr__(self)
        """
        return repr(self._create_list())

    def __str__(self):
        """
        __str__(self)
        """
        return str(self._create_list())


cdef class DOMNode:
    """
    __init__(self)

    DOM node.

    A DOM node is only valid as long as the owning :class:`HTMLTree` is alive
    and the DOM tree hasn't been modified. Do not access :class:`DOMNode` instances
    after any sort of DOM tree manipulation.
    """

    def __cinit__(self, HTMLTree tree):
        self.tree = tree
        self.node = NULL
        self.class_list_singleton = None

    def __dealloc__(self):
        if not self.node or self.tree is None:
            return

        self.node.user = NULL

        # Cannot be done until https://github.com/lexbor/lexbor/issues/132 is fixed
        # If you create lots of unparented DOMNodes, we may leak memory
        # if not self.node.parent and self.node != <lxb_dom_node_t*>self.node.owner_document:
        #     lxb_dom_node_destroy_deep(self.node)
        self.node = NULL

    def __iter__(self):
        """
        __iter__(self)

        Traverse the DOM tree in pre-order starting at the current node.

        :rtype: t.Iterable[DOMNode]
        """
        if not check_node(self):
            return

        yield self
        cdef lxb_dom_node_t* node = self.node
        while True:
            node = next_node(self.node, node)
            if not node:
                return
            yield _create_dom_node(self.tree, node)

    @property
    def type(self):
        """
        DOM node type.

        :type: NodeType
        """
        if not check_node(self):
            return None
        return <NodeType>self.node.type

    @property
    def tag(self):
        """
        DOM element tag or node name.

        :type: str or None
        """
        if not check_node(self):
            return None

        cdef size_t name_len = 0
        cdef const lxb_char_t* name = lxb_dom_node_name(self.node, &name_len)
        if not name:
            return None
        return name[:name_len].decode().lower()

    # noinspection DuplicatedCode
    @property
    def first_child(self):
        """
        First child element of this DOM node.

        :type: DOMNode or None
        """
        if not check_node(self):
            return None
        return _create_dom_node(self.tree, self.node.first_child)

    # noinspection DuplicatedCode
    @property
    def last_child(self):
        """
        Last child element of this DOM node.

        :type: DOMNode or None
        """
        if not check_node(self):
            return None
        return _create_dom_node(self.tree, self.node.last_child)

    # noinspection DuplicatedCode
    @property
    def first_element_child(self):
        """
        First element child of this DOM node.

        :type: DOMNode or None
        """
        if not check_node(self):
            return None
        cdef lxb_dom_node_t* child_node = self.node.first_child
        while child_node and child_node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            child_node = child_node.next
        return _create_dom_node(self.tree, child_node)

    # noinspection DuplicatedCode
    @property
    def last_element_child(self):
        """
        Last element child element of this DOM node.

        :type: DOMNode or None
        """
        if not check_node(self):
            return None
        cdef lxb_dom_node_t* child_node = self.node.last_child
        while child_node and child_node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            child_node = child_node.prev
        return _create_dom_node(self.tree, child_node)

    @property
    def child_nodes(self):
        """
        List of child nodes.

        :type: t.List[DOMNode]
        """
        if not check_node(self):
            return

        cdef lxb_dom_node_t* child = self.node.first_child
        child_nodes = []
        while child:
            child_nodes.append(_create_dom_node(self.tree, child))
            child = child.next
        return child_nodes

    @property
    def parent(self):
        """
        Parent of this node.

        :type: DOMNode or None
        """
        if not check_node(self):
            return None
        return _create_dom_node(self.tree, self.node.parent)

    # noinspection DuplicatedCode
    @property
    def next(self):
        """
        Next sibling node.

        :type: DOMNode or None
        """
        if not check_node(self):
            return None
        return _create_dom_node(self.tree, self.node.next)

    # noinspection DuplicatedCode
    @property
    def prev(self):
        """
        Previous sibling node.

        :type: DOMNode or None
        """
        if not check_node(self):
            return None
        return _create_dom_node(self.tree, self.node.prev)

    # noinspection DuplicatedCode
    @property
    def next_element(self):
        """
        Next sibling element node.

        :type: DOMNode or None
        """
        if not check_node(self):
            return None

        cdef lxb_dom_node_t* next_node = self.node.next
        while next_node and next_node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            next_node = next_node.next
        return _create_dom_node(self.tree, next_node)

    # noinspection DuplicatedCode
    @property
    def prev_element(self):
        """
        Previous sibling element node.

        :type: DOMNode or None
        """
        if not check_node(self):
            return None

        cdef lxb_dom_node_t* prev_node = self.node.prev
        while prev_node and prev_node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            prev_node = prev_node.prev
        return _create_dom_node(self.tree, prev_node)

    @property
    def value(self):
        """
        Node text value.

        :type: str or None
        """
        if self.node.type not in [LXB_DOM_NODE_TYPE_TEXT,
                                  LXB_DOM_NODE_TYPE_PROCESSING_INSTRUCTION,
                                  LXB_DOM_NODE_TYPE_COMMENT]:
            return None

        cdef lxb_dom_character_data_t* cd = <lxb_dom_character_data_t*>self.node
        return cd.data.data[:cd.data.length].decode()

    @property
    def text(self):
        """
        Text contents of this DOM node and its children.

        The DOM node's inner text can be modified by assigning to this property.

        :type: str
        """
        if not check_node(self):
            return None

        return get_node_text(self.node).decode()

    @text.setter
    def text(self, str text):
        if not check_node(self):
            raise RuntimeError('Trying to set text contents of uninitialized DOM node')

        cdef bytes text_bytes = text.encode()
        lxb_dom_node_text_content_set(self.node, <const lxb_char_t*>text_bytes, len(text_bytes))

    @property
    def html(self):
        """
        HTML contents of this DOM node and its children.

        The DOM node's inner HTML can be modified by assigning to this property.

        :type: str
        """
        if not check_node(self):
            return None

        cdef lexbor_str_t* html_str = lexbor_str_create()
        lxb_html_serialize_tree_str(self.node, html_str)
        cdef str py_text = html_str.data[:html_str.length].decode()
        lexbor_str_destroy(html_str, self.node.owner_document.text, True)
        return py_text

    @html.setter
    def html(self, str html):
        if not check_node(self):
            raise RuntimeError('Trying to set HTML contents of uninitialized DOM node')

        cdef bytes html_bytes = html.encode()
        lxb_html_element_inner_html_set(<lxb_html_element_t*>self.node, <const lxb_char_t*>html_bytes, len(html_bytes))

    @property
    def id(self):
        """
        ID attribute of this Element node (empty string if unset).

        :type: str
        """
        if not check_node(self) or self.node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            return None

        cdef size_t value_len = 0
        cdef const lxb_char_t* value = lxb_dom_element_id(<lxb_dom_element_t*>self.node, &value_len)
        if not value:
            return ''
        return value[:value_len].decode()

    @id.setter
    def id(self, str class_name):
        self._setattr_impl(b'id', class_name.encode())

    @property
    def class_name(self):
        """
        Class name attribute of this Element node (empty string if unset).

        :type: str
        """
        if not check_node(self) or self.node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            return None

        cdef size_t value_len = 0
        cdef const lxb_char_t* value = lxb_dom_element_class(<lxb_dom_element_t*>self.node, &value_len)
        if not value:
            return ''
        return value[:value_len].decode()

    @class_name.setter
    def class_name(self, str class_name):
        self._setattr_impl(b'class', class_name.encode())

    @property
    def class_list(self):
        """
        List of class names set on this Element node.

        :type: DOMElementClassList
        """
        if not check_node(self) or self.node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            return None

        if self.class_list_singleton is None:
            # noinspection PyAttributeOutsideInit
            self.class_list_singleton = DOMElementClassList.__new__(DOMElementClassList, self)
        return self.class_list_singleton

    @property
    def attrs(self):
        """
        List of attribute names if node is an Element node.

        :type: t.List[str] or None
        """
        if not check_node(self) or self.node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            return None

        cdef lxb_dom_attr_t* attr = lxb_dom_element_first_attribute(<lxb_dom_element_t*>self.node)
        cdef const lxb_char_t* local_name
        cdef size_t local_name_len = 0

        attrs = []
        while attr:
            local_name = lxb_dom_attr_local_name(attr, &local_name_len)
            attrs.append(local_name[:local_name_len].decode())
            attr = attr.next

        return attrs

    cpdef bint hasattr(self, str attr_name) except -1:
        """
        hasattr(self, attr_name)
        
        Check if node has an attribute with the given name.

        :param attr_name: attribute name
        :type attr_name: str
        :rtype: bool
        :raises ValueError: if node is not an Element node
        """
        if not check_node(self) or self.node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            raise ValueError('Node ist not an Element node.')

        cdef bytes attr_name_bytes = attr_name.encode()
        return <bint>lxb_dom_element_has_attribute(<lxb_dom_element_t*>self.node,
                                                   <const lxb_char_t*>attr_name_bytes, len(attr_name_bytes))

    cdef str _getattr_impl(self, const string& attr_name, bint raise_if_not_exists=True):
        """
        Get an attribute value as string.
        
        :param attr_name: 
        :return: attribute value or None
        :raises KeyError: if attribute does not exist
        """
        if not check_node(self) or self.node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            raise ValueError('Node ist not an Element node.')

        if raise_if_not_exists and not lxb_dom_element_has_attribute(
                <lxb_dom_element_t*>self.node, <const lxb_char_t*>attr_name.data(), attr_name.size()):
            raise KeyError(f'No such attribute: {attr_name.decode()}')

        cdef size_t node_attr_len
        cdef const lxb_char_t* node_attr_data = lxb_dom_element_get_attribute(
            <lxb_dom_element_t*>self.node, <const lxb_char_t*>attr_name.data(), attr_name.size(), &node_attr_len)
        return node_attr_data[:node_attr_len].decode()

    cpdef str getattr(self, str attr_name, str default_value=None):
        """
        getattr(self, attr_name, default_value=None)

        Get the value of the attribute ``attr_name`` or ``default_value`` if the element has no such attribute.

        :param attr_name: attribute name
        :type attr_name: str
        :param default_value: default value to return if attribute is unset
        :type default_value: str
        :return: attribute value
        :rtype: str
        :raises ValueError: if node is not an Element node
        """
        if not check_node(self):
            return None

        try:
            return self._getattr_impl(attr_name.encode())
        except KeyError:
            return default_value

    def __getitem__(self, str attr_name):
        """
        __getitem__(self, attr_name)

        Get the value of the an attribute.

        :param attr_name: attribute name
        :rtype: str
        :raises KeyError: if no such attribute exists
        :raises ValueError: if node is not an Element node
        """
        if not check_node(self):
            return None

        return self._getattr_impl(attr_name.encode())

    cdef bint _setattr_impl(self, bytes attr_name, bytes attr_value) except -1:
        """
        Insert or update an attribute with the given value.
        
        :param attr_name: attribute value
        :param attr_value: attribute name
        :return: attribute value or None
        """
        if not check_node(self) or self.node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            raise ValueError('Node ist not an Element node.')

        # Lexbor's check of existing attributes is buggy
        cdef lxb_dom_attr_t* attr = lxb_dom_element_attr_is_exist(<lxb_dom_element_t*>self.node,
                                                                  <const lxb_char_t*>attr_name, len(attr_name))

        if attr:
            lxb_dom_attr_set_value(attr, <const lxb_char_t*>attr_value, len(attr_value))
            return True

        lxb_dom_element_set_attribute(<lxb_dom_element_t*>self.node,
                                      <const lxb_char_t*>attr_name, len(attr_name),
                                      <const lxb_char_t*>attr_value, len(attr_value))
        return True

    cpdef setattr(self, str attr_name, str attr_value):
        """
        setattr(self, attr_name, attr_value)
        
        Insert or update an attribute with the given name to the given value.

        :param attr_name: attribute name
        :type attr_name: str
        :param attr_value: attribute value
        :type attr_value: str
        :return: attribute value
        :raises ValueError: if node is not an Element node
        """
        if not check_node(self):
            return

        self._setattr_impl(attr_name.encode(), attr_value.encode())

    def __setitem__(self, str attr_name, str attr_value):
        """
        __setitem__(self, attr_name, attr_value)

        Insert or update an attribute with the given name to the given value.

        :param attr_name: attribute name
        :type attr_name: str
        :param attr_value: attribute value
        :type attr_value: str
        :return: attribute value
        :raises ValueError: if node is not an Element node
        """
        if not check_node(self):
            return

        self._setattr_impl(attr_name.encode(), attr_value.encode())

    cdef bint _delattr_impl(self, bytes attr_name) except -1:
        """
        Remove the given attribute.

        :param attr_name: attribute to remove
        """
        if not check_node(self) or self.node.type != LXB_DOM_NODE_TYPE_ELEMENT:
            raise ValueError('Node ist not an Element node.')

        cdef lxb_dom_attr_t* attr = lxb_dom_element_attr_is_exist(<lxb_dom_element_t*>self.node,
                                                                  <const lxb_char_t*>attr_name, len(attr_name))
        if not attr:
            return False

        lxb_dom_element_attr_remove(<lxb_dom_element_t*>self.node, attr)
        return True

    cpdef delattr(self, str attr_name):
        """
        delattr(self, attr_name)
        
        Remove the given attribute if it exists.

        :param attr_name: attribute to remove
        :type attr_name: str
        :raises ValueError: if node is not an Element node
        """
        self._delattr_impl(attr_name.encode())

    def __delitem__(self, str attr_name):
        """
        __delitem__(self, attr_name)

        Remove the given attribute.

        :param attr_name: attribute to remove
        :type attr_name: str
        :raises ValueError: if node is not an Element node
        :raises KeyError: if node 
        """
        cdef bint ret = self._delattr_impl(attr_name.encode())
        if not ret:
            raise KeyError(f'No such attribute: {attr_name}')

    cpdef DOMNode query_selector(self, str selector):
        """
        query_selector(self, selector)
         
        Find and return the first element matching the given CSS selector. This is more efficient
        than matching with :meth:`query_selector_all` and discarding additional elements.

        :param selector: CSS selector
        :type selector: str
        :return: matching element or ``None``
        :rtype: DOMNode or None
        """
        cdef lxb_dom_node_t* node = query_selector_impl(self.node, self.tree, selector.encode())
        if not node:
            return None
        return _create_dom_node(self.tree, node)

    cpdef DOMCollection query_selector_all(self, str selector):
        """
        query_selector_all(self, selector)
        
        Find all elements matching the given CSS selector and return a :class:`DOMCollection` with the results.

        :param selector: CSS selector
        :type selector: str
        :return: collection of matching elements
        :rtype: DOMCollection
        """
        cdef lxb_dom_collection_t* coll = query_selector_all_impl(self.node, self.tree, selector.encode())
        if not coll:
            raise RuntimeError('Failed to match elements by CSS selector')

        return _create_dom_collection(self.tree, coll)

    cpdef bint matches(self, str selector) except -1:
        """
        matches(self, selector)
        
        Check whether any element in the DOM tree matches the given CSS selector. This is more efficient
        than matching with :meth:`query_selector_all` and checking the size of the returned collection.

        :param selector: CSS selector
        :type selector: str
        :return: boolean value indicating whether a matching element exists
        :rtype: bool
        """
        self.tree.init_css_parser()
        return matches_impl(self.node, self.tree, selector.encode())

    cpdef DOMCollection get_elements_by_attr(self, str attr_name, str attr_value, bint case_insensitive=False):
        """
        get_elements_by_attr(self, attr_name, attr_value, case_insensitive=False)
        
        Find all elements matching the given arbitrary attribute name and value and return a
        :class:`DOMCollection` with the results.
        
        :param attr_name: attribute name
        :type attr_name: str
        :param attr_value: attribute value
        :type attr_name: str
        :param case_insensitive: match attribute value case-insensitively
        :type case_insensitive: bool
        :return: collection of matching elements
        :rtype: DOMCollection or None
        """
        if not check_node(self):
            return None

        return _create_dom_collection(self.tree,
                                      get_elements_by_attr_impl(self.node, attr_name.encode(), attr_value.encode(),
                                                                10, case_insensitive))

    cpdef DOMNode get_element_by_id(self, str element_id, bint case_insensitive=False):
        """
        get_element_by_id(self, element_id, case_insensitive=False)
        
        Find and return the element whose ID attribute matches ``element_id``.
        
        :param element_id: element ID
        :type element_id: str
        :param case_insensitive: match ID case-insensitively
        :type case_insensitive: bool
        :return: matching element or ``None`` if no such element exists
        :rtype: DOMNode or None
        """
        if not check_node(self):
            return None

        cdef lxb_dom_node_t* result_node = get_element_by_id_impl(self.node, element_id.encode(), case_insensitive)
        if not result_node:
            return None
        return _create_dom_node(self.tree, result_node)

    cpdef DOMCollection get_elements_by_class_name(self, str class_name, bint case_insensitive=False):
        """
        get_elements_by_class_name(self, element_class, case_insensitive=False)
        
        Find all elements matching the given class name and return a :class:`DOMCollection` with the results.
        
        :param class_name: element class
        :type class_name: str
        :param case_insensitive: match class name case-insensitively
        :type case_insensitive: bool
        :return: collection of matching elements
        :rtype: DOMCollection or None
        """
        if not check_node(self):
            return None

        return _create_dom_collection(self.tree, get_elements_by_class_name_impl(self.node, class_name.encode(), 10))

    cpdef DOMCollection get_elements_by_tag_name(self, str tag_name):
        """
        get_elements_by_tag_name(self, tag_name)
        
        Find all elements with the given tag name and return a :class:`DOMCollection` with the results.
        
        :param tag_name: tag name for matching elements
        :type tag_name: str
        :return: collection of matching elements
        :rtype: DOMCollection
        """
        if not check_node(self):
            return None

        return _create_dom_collection(self.tree, get_elements_by_tag_name_impl(self.node, tag_name.encode()))

    cpdef DOMNode append_child(self, DOMNode node):
        """
        append_child(self, node)
        
        Append a new child node to this DOM node.
        
        :param node: DOM node to append as new child node
        :type node: DOMNode
        :return: the appended child node
        :rtype: DOMNode
        :raises ValueError: if trying to append node to itself
        """
        if not check_node(self) or not check_node(node):
            raise RuntimeError('Append operation on uninitialized node')

        if node.node == self.node:
            raise ValueError('Trying to append child to itself')

        if node.node.parent:
            lxb_dom_node_remove(node.node)
        lxb_dom_node_insert_child(self.node, node.node)
        return node

    cpdef DOMNode insert_before(self, DOMNode node, DOMNode reference):
        """
        insert_before(self, node, reference)
        
        Insert ``node`` before ``reference`` as a new child node. The reference node must be
        a child of this node or ``None``. If ``reference`` is ``None``, the new node
        will be appended after the last child node.
        
        :param node: DOM node to insert as new child node
        :type node: DOMNode
        :param reference: child node before which to insert the new node or ``None``
        :type reference: DOMNode
        :return: the inserted child node
        :rtype: DOMNode
        :raises ValueError: if trying to add node as its own child or if ``reference`` is not a child
        """
        if not check_node(self) or not check_node(node) or not check_node(reference):
            raise RuntimeError('Insert operation on uninitialized node')

        if node.node == self.node:
            raise ValueError('Trying to insert node as its own child')

        if reference.node.parent != self.node:
            raise ValueError('Reference node must be a child node')

        if node.node.parent:
            lxb_dom_node_remove(node.node)
        lxb_dom_node_insert_before(reference.node, node.node)
        return node

    cpdef DOMNode replace_child(self, DOMNode new_child, DOMNode old_child):
        """
        replace_child(self, new_child, old_child)
        
        Replace the child node ``old_child`` with ``new_child``.
        
        :param new_child: new child node to insert
        :type new_child: DOMNode
        :param old_child: old child node to replace
        :type old_child: DOMNode
        :return: the old child node
        :rtype: DOMNode
        :raises ValueError: if ``old_child`` is not a child of this node
        """
        if not check_node(self) or not check_node(new_child) or not check_node(old_child):
            raise RuntimeError('Replace operation on uninitialized node')

        if old_child.node.parent != self.node:
            raise ValueError('Node is not a child of the current node')

        if new_child.node == old_child.node:
            return old_child

        if new_child.node.parent:
            lxb_dom_node_remove(new_child.node)
        lxb_dom_node_insert_after(old_child.node, new_child.node)
        lxb_dom_node_remove(old_child.node)

        return old_child

    cpdef DOMNode remove_child(self, DOMNode node):
        """
        remove_child(self, node)
        
        Remove the child node ``node`` from the DOM tree and return it.
        
        :param node: DOM node to remove
        :type node: DOMNode
        :return: the removed child node
        :rtype: DOMNode
        :raises ValueError: if ``node`` is not a child of this node
        """
        if not check_node(self) or not check_node(node):
            raise RuntimeError('Remove operation on uninitialized node')

        if node.node.parent != self.node:
            raise ValueError('Node is not a child of the current node')

        lxb_dom_node_remove(node.node)
        return node

    # noinspection PyAttributeOutsideInit
    cpdef decompose(self):
        """
        decompose(self)
        
        Delete the current node and all its children.
        """
        if not check_node(self):
            raise RuntimeError('Decompose operation on uninitialized node')

        lxb_dom_node_destroy_deep(self.node)
        self.node = NULL
        self.tree = None

    def __repr__(self):
        if not check_node(self):
            return '<INVALID ELEMENT>'

        if self.node.type == LXB_DOM_NODE_TYPE_ELEMENT:
            attrs = ' '.join(f'{a}="{self[a]}"' for a in self.attrs)
            if attrs:
                attrs = ' ' + attrs
            return f'<{self.tag}{attrs}>'
        elif self.node.type == LXB_DOM_NODE_TYPE_TEXT:
            return self.text
        elif self.node.type == LXB_DOM_NODE_TYPE_DOCUMENT:
            return '[HTML Document]'
        elif self.node.type == LXB_DOM_NODE_TYPE_DOCUMENT_TYPE:
            return '<!DOCTYPE html>'

        return f'<{self.__class__.__name__} Element>'

    def __str__(self):
        return self.html

    def __eq__(self, other):
        if not check_node(self) or not isinstance(other, DOMNode):
            return False
        cdef DOMNode other_dom = <DOMNode>other
        return other_dom.node == self.node

    def __hash__(self):
        if not check_node(self):
            return 0
        return <size_t>self.node


cdef inline void _join_collections(lxb_dom_collection_t* target, lxb_dom_collection_t* source):
    """
    Append elements from ``source` to ``target``.
    
    :param target: target collection to append to
    :param source: source collection to append from
    """

    cdef size_t i = 0
    for i in range(lxb_dom_collection_length(source)):
        lxb_dom_collection_append(target, lxb_dom_collection_element(source, i))


cdef class DOMCollection:
    """
    __init__(self)

    Collection of DOM nodes that are the result set of an element matching operation.

    A node collection is only valid for as long as the owning :class:`HTMLTree` is alive
    and the DOM tree hasn't been modified. Do not access :class:`DOMCollection` instances
    after any sort of DOM tree manipulation.
    """

    def __cinit__(self, HTMLTree tree):
        self.tree = tree
        self.coll = NULL

    def __dealloc__(self):
        if self.coll:
            lxb_dom_collection_destroy(self.coll, True)
            self.coll = NULL

    @cython.cdivision(False)
    cdef inline size_t _wrap_idx(self, ssize_t idx):
        if idx >= 0:
            return idx
        return idx % <ssize_t>lxb_dom_collection_length(self.coll)

    cdef DOMCollection _forward_collection_match(self, bytes func, attrs):
        """
        Forward DOM multi-element match operations to all items in the collection and aggregate the results.
        
        :param func: internal matching function as bytes (b'by_attr', b'by_class', b'by_tag', or b'selector_all')
        :param attrs: tuple of attributes to pass to the matching function
        :return: aggregated collection or single element
        """
        if self.tree is None or not self.coll:
            raise RuntimeError('Trying to select items from uninitialized collection')

        if not lxb_dom_collection_length(self.coll):
            return self

        cdef lxb_dom_collection_t* joined_coll = lxb_dom_collection_make(self.coll.document,
                                                                         lxb_dom_collection_length(self.coll) * 2)

        cdef lxb_dom_node_t* node = NULL
        cdef lxb_dom_collection_t* matches = NULL
        cdef size_t i = 0
        for i in range(lxb_dom_collection_length(self.coll)):
            node = lxb_dom_collection_node(self.coll, i)
            if func == b'by_attr':
                matches = get_elements_by_attr_impl(node, attrs[0], attrs[1], attrs[2], attrs[3])
            if func == b'by_class':
                matches = get_elements_by_class_name_impl(node, attrs[0], attrs[1])
            elif func == b'by_tag':
                matches = get_elements_by_tag_name_impl(node, attrs[0])
            elif func == b'selector_all':
                matches = query_selector_all_impl(node, self.tree, attrs[0])

            if matches:
                _join_collections(joined_coll, matches)
                lxb_dom_collection_destroy(matches, True)

        return _create_dom_collection(self.tree, joined_coll)

    cpdef DOMNode get_element_by_id(self, str element_id, bint case_insensitive=False):
        """
        get_element_by_id(self, element_id, case_insensitive=False)

        Within all elements in this collection, find and return the element whose ID
        attribute matches ``element_id``.

        :param element_id: element ID
        :type element_id: str
        :param case_insensitive: match ID case-insensitively
        :type case_insensitive: bool
        :return: matching element or ``None`` if no such element exists
        :rtype: DOMNode or None
        """
        if self.tree is None or not self.coll:
            raise RuntimeError('Trying to select items from uninitialized collection')

        cdef lxb_dom_node_t* node
        cdef size_t i = 0
        cdef bytes id_bytes = element_id.encode()
        for i in range(lxb_dom_collection_length(self.coll)):
            node = get_element_by_id_impl(lxb_dom_collection_node(self.coll, i), id_bytes, case_insensitive)
            if node:
                return _create_dom_node(self.tree, node)

        return None

    cpdef DOMCollection get_elements_by_attr(self, str attr_name, str attr_value, bint case_insensitive=False):
        """
        get_elements_by_attr(self, attr_name, attr_value, case_insensitive=False)

        Within all elements in this collection, find the elements matching the given arbitrary attribute
        name and value and return a :class:`DOMCollection` with the aggregated results.

        :param attr_name: attribute name
        :type attr_name: str
        :param attr_value: attribute value
        :type attr_name: str
        :param case_insensitive: match attribute value case-insensitively
        :type case_insensitive: bool
        :return: collection of matching elements
        :rtype: DOMCollection or None
        """
        return self._forward_collection_match(b'by_attr',
                                              (attr_name.encode(), attr_value.encode(),  10, case_insensitive))

    cpdef DOMCollection get_elements_by_class_name(self, str class_name, bint case_insensitive=False):
        """
        get_elements_by_class_name(self, element_class, case_insensitive=False)

        Within all elements in this collection, find the elements matching the given class name
        and return a :class:`DOMCollection` with the aggregated results.

        :param class_name: element class
        :type class_name: str
        :param case_insensitive: match class name case-insensitively
        :type case_insensitive: bool
        :return: collection of matching elements
        :rtype: DOMCollection or None
        """
        return self._forward_collection_match(b'by_class', (class_name.encode(), 10))

    cpdef DOMCollection get_elements_by_tag_name(self, str tag_name):
        """
        get_elements_by_tag_name(self, tag_name)

        Within all elements in this collection, find the elements with the given tag name and return
        a :class:`DOMCollection` with the aggregated results.

        :param tag_name: tag name for matching elements
        :type tag_name: str
        :return: collection of matching elements
        :rtype: DOMCollection
        """
        return self._forward_collection_match(b'by_tag', (tag_name.encode(),))

    cpdef DOMNode query_selector(self, str selector):
        """
        query_selector(self, selector)

        Within all elements in this collection, find and return the first element matching
        the given CSS selector.This is more efficient than matching with :meth:`query_selector_all`
        and discarding additional elements.

        :param selector: CSS selector
        :type selector: str
        :return: matching element or ``None``
        :rtype: DOMNode or None
        """
        if self.tree is None or not self.coll:
            raise RuntimeError('Trying to select items from uninitialized collection')

        cdef lxb_dom_node_t* node
        cdef size_t i = 0
        cdef bytes selector_bytes = selector.encode()
        for i in range(lxb_dom_collection_length(self.coll)):
            node = query_selector_impl(lxb_dom_collection_node(self.coll, i), self.tree, selector_bytes)
            if node:
                return _create_dom_node(self.tree, node)

        return None

    cpdef DOMCollection query_selector_all(self, str selector):
        """
        query_selector_all(self, selector)

        Within all elements in this collection, find the elements matching the given CSS selector
        and return a :class:`DOMCollection` with the aggregated results.

        :param selector: CSS selector
        :type selector: str
        :return: collection of matching elements
        :rtype: DOMCollection
        """
        return self._forward_collection_match(b'selector_all', (selector.encode(),))

    cpdef bint matches(self, str selector) except -1:
        """
        matches(self, selector)

        Within all elements in this collection, check whether any element in the DOM tree
        matches the given CSS selector. This is more efficient than matching with
        :meth:`query_selector_all` and checking the size of the returned collection.

        :param selector: CSS selector
        :type selector: str
        :return: boolean value indicating whether a matching element exists
        :rtype: bool
        """
        if self.tree is None or not self.coll:
            raise RuntimeError('Trying to select items from uninitialized collection')

        cdef size_t i = 0
        cdef bytes selector_bytes = selector.encode()
        for i in range(lxb_dom_collection_length(self.coll)):
            if matches_impl(lxb_dom_collection_node(self.coll, i), self.tree, selector_bytes):
                return True

        return False

    def __iter__(self):
        """
        __iter__(self)

        Iterate the DOM node collection.

        :rtype: t.Iterable[DOMNode]
        """
        if self.tree is None or not self.coll:
            raise RuntimeError('Trying to get item of uninitialized collection')

        cdef size_t i = 0
        for i in range(lxb_dom_collection_length(self.coll)):
            yield _create_dom_node(self.tree, lxb_dom_collection_node(self.coll, i))

    def __len__(self):
        """
        __len__(self)

        Get the number of items in this collection.

        :rtype: int
        """
        if self.tree is None or not self.coll:
            return 0
        return lxb_dom_collection_length(self.coll)

    def __getitem__(self, key):
        """
        __getitem__(self, key)

        Return the :class:`DOMNode` at the given index in this collection or another :class:`DOMCollection`
        if ``key`` is a slice object. Negative indexing is supported.

        :param key: index or slice
        :rtype: DOMNode or DOMCollection
        :raises IndexError: if ``key`` is out of range
        :raises TypeError: if ``key`` is not an ``int`` or ``slice``
        """
        if self.tree is None or not self.coll:
            raise IndexError('Trying to get item of uninitialized collection')

        cdef size_t coll_len = lxb_dom_collection_length(self.coll)

        cdef DOMCollection slice_coll
        cdef lxb_dom_collection_t* dom_coll
        if isinstance(key, slice):
            start = key.start
            stop = key.stop
            step = key.step if key.step is not None else 1

            if start is None:
                start = coll_len - 1 if step < 0 else 0
            else:
                start = self._wrap_idx(min(start, coll_len - 1))

            if stop is None:
                stop = -1 if step < 0 else coll_len
            else:
                stop = self._wrap_idx(min(stop, coll_len))

            dom_coll = lxb_dom_collection_make(self.coll.document,
                                               min(coll_len, abs((stop - start) // step) + 1))
            for i in range(start, stop, step):
                lxb_dom_collection_append(dom_coll, lxb_dom_collection_element(self.coll, i))

            slice_coll = DOMCollection.__new__(DOMCollection, self.tree)
            slice_coll.coll = dom_coll
            return slice_coll

        if type(key) is not int:
            raise TypeError(f'Invalid key type: {type(key)}')

        if key >= coll_len:
            raise IndexError('Index out of range')

        return _create_dom_node(self.tree, lxb_dom_collection_node(self.coll, self._wrap_idx(key)))

    def __repr__(self):
        return f'{{{", ".join(repr(n) for n in self)}}}'

    def __str__(self):
        return self.__repr__()


cdef HTMLTree create_html_tree(bytes document, bint reencode=True, str encoding='utf-8', str errors='ignore'):
    """
    Create :class:`HTMLTree` instance from bytes content.
    
    :param document: input document bytes
    :param reencode: whether to reencode the bytes to erase encoding errors
    :param encoding: input encoding (best guess)
    :param errors: decoding error policy
    :return: :class:`HTMLTree` instance
    """
    encoding = map_encoding_to_html5(encoding)
    if reencode:
        document = bytes_to_str(document, encoding, errors).encode()
    cdef lxb_status_t status
    cdef const lxb_char_t* html = <const lxb_char_t*>document
    cdef size_t html_len = len(document)
    cdef HTMLTree tree = HTMLTree.__new__(HTMLTree)
    with nogil:
        status = lxb_html_document_parse(tree.dom_document, html, html_len)
    if status != LXB_STATUS_OK:
        raise ValueError('Failed to parse HTML document')
    return tree


cdef class HTMLTree:
    """
    __init__(self)

    HTML DOM tree parser.
    """

    def __init__(self, *args, **kwargs):
        raise SyntaxError('Cannot create instance directly. Use HTMLParser.parse() or HTMLParser.parse_from_bytes()')

    def __cinit__(self):
        self.dom_document = lxb_html_document_create()
        if not self.dom_document:
            raise RuntimeError('Failed to allocate HTML document')
        self.css_parser = NULL
        self.css_selectors = NULL
        self.selectors = NULL

    def __dealloc__(self):
        if self.dom_document:
            lxb_html_document_destroy(self.dom_document)
            self.dom_document = NULL

        if self.css_parser or self.css_parser or self.css_selectors:
            destroy_css_selectors(self.css_selectors, self.selectors)
            destroy_css_parser(self.css_parser)
            self.css_parser = NULL
            self.css_parser = NULL
            self.css_selectors = NULL

    # noinspection PyAttributeOutsideInit
    cdef inline void init_css_parser(self):
        """
        Initialize CSS selector if not already initialized.
        """
        if not self.css_parser:
            init_css_parser(&self.css_parser)
            init_css_selectors(self.css_parser, &self.css_selectors, &self.selectors)


    @classmethod
    def parse(cls, str document):
        """
        parse(self, document)
        
        Parse HTML from a Unicode string into a DOM tree.
        
        :param document: input HTML document
        :return: HTML DOM tree
        :rtype: HTMLTree
        :raises ValueError: if HTML parsing fails for unknown reasons
        """
        return create_html_tree(document.encode(), reencode=False)

    @classmethod
    def parse_from_bytes(cls, bytes document, str encoding='utf-8', str errors='ignore'):
        """
        parse_from_bytes(self, document, encoding='utf-8', errors='ignore')
        
        Decode a raw HTML byte string and parse it into a DOM tree.
        
        The decoding routine uses :func:`~.parse.encoding.bytes_to_str` to take care of decoding errors,
        so it is sufficient if ``encoding`` is just a best guess of what the actual input encoding is.
        The encoding name will be remapped according to the WHATWG specification by calling
        :func:`~.parse.encoding.map_encoding_to_html5` before trying to decode the byte string with it.
        
        :param document: input byte string
        :param encoding: encoding for decoding byte string
        :param errors: decoding error policy (same as ``str.decode()``)
        :return: HTML DOM tree
        :rtype: HTMLTree
        :raises ValueError: if HTML parsing fails for unknown reasons
        """
        return create_html_tree(document, True, encoding, errors)

    @property
    def document(self):
        """
        Document root node.

        :type: DOMNode or None
        """
        if not self.dom_document:
            return None

        return _create_dom_node(self, <lxb_dom_node_t*>&self.dom_document.dom_document)

    @property
    def head(self):
        """
        HTML head element or ``None`` if document has no head.

        :type: DOMNode or None
        """
        if not self.dom_document:
            return None

        return _create_dom_node(self, <lxb_dom_node_t*>lxb_html_document_head_element(self.dom_document))

    @property
    def body(self):
        """
        HTML body element or ``None`` if document has no body.

        :type: DOMNode or None
        """
        if not self.dom_document:
            return None

        return _create_dom_node(self, <lxb_dom_node_t*>lxb_html_document_body_element(self.dom_document))

    @property
    def title(self):
        """
        The HTML document title.

        :type: str or None
        """
        if not self.dom_document or not self.head:
            return None

        cdef size_t title_len = 0
        cdef const lxb_char_t* title = lxb_html_document_title(self.dom_document, &title_len)
        if not title:
            return ''
        return title[:title_len].decode()

    cpdef DOMNode create_element(self, str tag_name):
        """
        create_element(self, tag_name)
        
        Create a new DOM Element node.
        
        :param tag_name: element tag name
        :type tag_name: str
        :return: new Element node
        :rtype: DOMNode
        """
        if not self.dom_document:
            raise RuntimeError('Trying to create element in uninitialized document.')

        cdef bytes tag_name_bytes = tag_name.encode()
        cdef lxb_dom_element_t* element = lxb_dom_document_create_element(
            <lxb_dom_document_t*>self.dom_document, <const lxb_char_t*>tag_name_bytes, len(tag_name_bytes), NULL)
        return _create_dom_node(self, <lxb_dom_node_t*>element)

    cpdef DOMNode create_text_node(self, str text):
        """
        create_text_node(self, text)

        Create a new DOM Element node.

        :param text: string contents of the new text element
        :type text: str
        :return: new text node
        :rtype: DOMNode
        """
        if not self.dom_document:
            raise RuntimeError('Trying to create text node in uninitialized document.')

        cdef bytes text_bytes = text.encode()
        cdef lxb_dom_text_t* node = lxb_dom_document_create_text_node(
            <lxb_dom_document_t*>self.dom_document, <const lxb_char_t*>text_bytes, len(text_bytes))
        return _create_dom_node(self, <lxb_dom_node_t*>node)

    def __str__(self):
        cdef DOMNode doc = self.document
        if doc is not None:
            return doc.html
        return ''


class DOMContext:
    """
    __init__()

    DOM node traversal context object.

    The context object has two attributes that are set by the traversal function for
    keeping track of the current :class:`DOMNode` and the current traversal depth.
    Besides these, the context object is arbitrarily mutable and can be used for
    maintaining custom state.

    :ivar DOMNode node: the current :class:`DOMNode`
    :ivar int depth: the current traversal depth
    """
    def __init__(self):
        self.node = None
        self.depth = 0


def traverse_dom(DOMNode base_node, start_callback, end_callback=None, context=None, bint elements_only=False):
    """
    traverse_dom(base_node, start_callback, end_callback=None, context=None, elements_only=False)

    DOM traversal helper.

    Traverses the DOM tree starting at ``base_node`` in pre-order and calls ``start_callback``
    at each child node. If ``end_callback`` is not ``None``, it will be called each time
    a DOM element's end tag is encountered.

    The callbacks are expected to take exactly one :class:`DOMContext` context parameter,
    which keeps track of the current node and traversal depth. The context object will be
    the same throughout the whole traversal process, so it can be mutated with custom data.

    :param base_node: root node of the traversal
    :type base_node: DOMNode
    :param start_callback: callback for each DOM node on the way (takes a :class:`DOMNode`
                           and ``context`` as a parameter)
    :type start_callback: t.Callable[[DOMContext], None]
    :param end_callback: optional callback for element node end tags (takes a :class:`DOMNode`
                         and ``context`` as a parameter)
    :type end_callback: t.Callable[[DOMContext], None] or None
    :param context: optional pre-initialized context object
    :type context: DOMContext
    :param elements_only: traverse only element nodes
    :type elements_only: bool
    """

    cdef lxb_dom_node_t* node = base_node.node
    cdef size_t depth = 0
    cdef bint is_end_tag = False
    cdef bint* is_end_tag_ptr = &is_end_tag if end_callback is not None else NULL

    if elements_only and base_node.node.type != LXB_DOM_NODE_TYPE_ELEMENT:
        return

    context = context or DOMContext()

    while node:
        context.node = _create_dom_node(base_node.tree, node)
        context.depth = depth

        if not is_end_tag:
            start_callback(context)
        else:
            end_callback(context)

        if elements_only:
            node = next_element_node(base_node.node, node, &depth, is_end_tag_ptr)
        else:
            node = next_node(base_node.node, node, &depth, is_end_tag_ptr)


cdef unordered_set[lxb_tag_id_t] BLOCK_ELEMENT_SET

cdef inline void _init_block_element_set() nogil:
    if not BLOCK_ELEMENT_SET.empty():
        return
    cdef size_t i
    for i in range(NUM_BLOCK_ELEMENTS):
        BLOCK_ELEMENT_SET.insert(BLOCK_ELEMENTS[i])

_init_block_element_set()


cdef inline bint is_block_element(lxb_tag_id_t tag_id) nogil:
    """
    Check whether an element is a block-level element.
    """
    return BLOCK_ELEMENT_SET.find(tag_id) != BLOCK_ELEMENT_SET.end()



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

import typing as t

from libcpp.string cimport string
from resiliparse_inc.cstdlib cimport strtol


cpdef bytes read_http_chunk(reader):
    """
    read_http_chunk(reader)

    Helper function for reading chunked HTTP payloads.

    Each call to this function will try to read the next chunk. In case of an error
    or EOF, an empty byte string will be returned.

    :param reader: input reader
    :type reader: fastwarc.stream_io.BufferedReader
    :return: contents of the next chunk or empty string if EOF
    :rtype: bytes
    """
    cdef string header_line = reader.readline()
    cdef size_t chunk_size = strtol(header_line.substr(0, header_line.size() - 2).c_str(), NULL, 16)
    return reader.read(chunk_size + 2)[:chunk_size]


def iterate_http_chunks(reader):
    """
    iterate_http_chunks(reader)

    Generator wrapping :func:`read_http_chunk` for fully consuming a chunked HTTP payload.

    :param reader: input reader
    :type reader: fastwarc.stream_io.BufferedReader
    :return: generator of chunks
    :rtype: t.Generator[bytes]
    """
    cdef bytes chunk
    while True:
        chunk = read_http_chunk(reader)
        if len(chunk) == 0:
            return
        yield chunk



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform
import os
import warnings
import sys

import distutils.ccompiler
from distutils.dir_util import copy_tree
from setuptools import find_packages, setup, Extension


# 1. BOILERPLATE -------------------------------------------------------

VERSION = '0.10.6'
ROOT_DIRECTORY = os.path.abspath(os.path.dirname(__file__))
CXX = distutils.ccompiler.get_default_compiler()

TRACE = bool(int(os.getenv('TRACE', 0)))
DEBUG = bool(int(os.getenv('DEBUG', 0))) or TRACE
ASAN = bool(int(os.getenv('ASAN', 0)))

USE_CYTHON = False
ext = 'cpp'
cython_args = {}
try:
    from Cython.Build import cythonize
    import Cython.Compiler.Options

    ext = 'pyx'
    cython_args = dict(
        annotate=DEBUG,
        language_level='3',
        compiler_directives=dict(
            linetrace=TRACE,
            initializedcheck=DEBUG,
            boundscheck=DEBUG,
            cdivision=True
        )
    )
    USE_CYTHON = True
except ModuleNotFoundError as e:
    pass

cpp_args = {}
if TRACE:
    cpp_args.update(dict(define_macros=[('CYTHON_TRACE_NOGIL', '1')]))

if CXX == 'unix':
    cpp_args.update(dict(
        extra_compile_args=['-std=c++17', '-O3', '-Wno-deprecated-declarations',
                            '-Wno-unreachable-code', '-Wno-unused-function',
                            # Temporary flags until https://github.com/lexbor/lexbor/pull/125 and
                            # https://github.com/lexbor/lexbor/pull/135 are released
                            '-fpermissive', '-Wno-c++11-narrowing'],
        extra_link_args=['-std=c++17']
    ))
    if ASAN:
        cpp_args['extra_compile_args'].append('-fsanitize=address')
        cpp_args['extra_link_args'].append('-fsanitize=address')

elif CXX == 'msvc':
    cpp_args.update(dict(
        extra_compile_args=['/std:c++latest'],
        extra_link_args=[]
    ))

if os.path.isdir(os.path.join(ROOT_DIRECTORY, '..', 'resiliparse_inc')):
    copy_tree(os.path.join(ROOT_DIRECTORY, '..', 'resiliparse_inc'),
              os.path.join(ROOT_DIRECTORY, 'resiliparse_inc'), update=1)
    copy_tree(os.path.join(ROOT_DIRECTORY, '..', 'resiliparse_common'),
              os.path.join(ROOT_DIRECTORY, 'resiliparse_common'), update=1)

data_ext = ['*.pxd', '*.h']
inc_package = []
if 'sdist' in sys.argv:
    # Include resiliparse_inc module and Cython src files only in source distribution
    data_ext.extend(['*.pyx', '*.pxi'])
    inc_package.append('resiliparse_inc')
    inc_package.append('resiliparse_common')


# 2. RESILIPARSE SETUP -------------------------------------------------------

fast_warc_src = os.path.abspath(os.path.join(ROOT_DIRECTORY, '..', 'fastwarc'))
if os.path.isdir(fast_warc_src):
    sys.path.insert(0, fast_warc_src)

resiliparse_extensions = [
    Extension('resiliparse.extract.html2text',
              sources=[f'resiliparse/extract/html2text.{ext}'], libraries=['lexbor', 're2'], **cpp_args),
    Extension('resiliparse.parse.encoding',
              sources=[f'resiliparse/parse/encoding.{ext}'], libraries=['uchardet', 'lexbor'], **cpp_args),
    Extension('resiliparse.parse.html',
              sources=[f'resiliparse/parse/html.{ext}'], libraries=['lexbor'], **cpp_args),
    Extension('resiliparse.parse.http',
              sources=[f'resiliparse/parse/http.{ext}'], **cpp_args),
    Extension('resiliparse.parse.lang',
              sources=[f'resiliparse/parse/lang.{ext}'], **cpp_args)
]
if os.name == 'posix':
    resiliparse_extensions.extend([
        Extension('resiliparse.process_guard', sources=[f'resiliparse/process_guard.{ext}'],
                  libraries=['pthread'], **cpp_args),
        Extension('resiliparse.itertools', sources=[f'resiliparse/itertools.{ext}'], **cpp_args)
    ])
else:
    warnings.warn(f"Unsupported platform '{platform.system()}': Building without ProcessGuard extension.")

if USE_CYTHON:
    resiliparse_extensions = cythonize(resiliparse_extensions, **cython_args)


setup(
    name='Resiliparse',
    version=VERSION,
    description='A collection of robust and fast processing tools for parsing and '
                'analyzing (not only) web archive data.',
    long_description=open(os.path.join(ROOT_DIRECTORY, 'README.md')).read(),
    long_description_content_type='text/markdown',
    author='Janek Bevendorff',
    url='https://github.com/chatnoir-eu/chatnoir-resiliparse',
    license='Apache License 2.0',
    packages=[*find_packages(), *inc_package],
    package_data={'': data_ext},
    ext_modules=resiliparse_extensions,
    install_requires=[
        'click',
        'fastwarc',
        'tqdm'
    ],
    setup_requires=['setuptools>=18.0'],
    tests_require=[
        'pytest',
        'pytest-cov'
    ],
    extras_require={
        "CLI": [
            'joblib'
        ],
    },
    entry_points={
        'console_scripts': ['resiliparse=resiliparse.cli:main[CLI]']
    }
)



# Copyright 2021 Janek Bevendorff
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# distutils: language = c++

from cython.operator cimport dereference as deref, preincrement as preinc, predecrement as predec
from libc.stdint cimport uint32_t
from libcpp.set cimport set as stl_set
from libc.string cimport memcpy
from libcpp.memory cimport make_shared, shared_ptr
from libcpp.string cimport string, to_string
from libcpp.vector cimport vector

from resiliparse.parse.encoding import bytes_to_str
from resiliparse_common.string_util cimport rstrip_str, strip_str
from resiliparse_inc.cctype cimport isspace
from resiliparse.parse.html cimport *
from resiliparse_inc.lexbor cimport *
from resiliparse_inc.re2 cimport Options as RE2Options, RE2Stack as RE2, StringPiece, PartialMatch


cdef extern from * nogil:
    """
    struct ExtractOpts {
        bool preserve_formatting = true;
        bool list_bullets = true;
        bool links = false;
        bool alt_texts = true;
        bool form_fields = false;
        bool noscript = false;
    };

    struct ExtractContext {
        lxb_dom_node_t* root_node = NULL;
        lxb_dom_node_t* node = NULL;
        size_t depth = 0;
        ExtractOpts opts;
    };

    struct ExtractNode {
        lxb_dom_node_t* reference_node = NULL;
        lxb_tag_id_t tag_id = LXB_TAG__UNDEF;
        size_t depth = 0;
        bool space_after = false;
        bool collapse_margins = true;
        bool is_big_block = false;
        bool is_pre = false;
        bool is_end_tag = false;
        std::shared_ptr<std::string> text_contents = NULL;
    };
    """
    cdef struct ExtractOpts:
        bint preserve_formatting
        bint list_bullets
        bint links
        bint alt_texts
        bint form_fields
        bint noscript

    cdef struct ExtractContext:
        lxb_dom_node_t * root_node
        lxb_dom_node_t * node
        size_t depth
        ExtractOpts opts

    cdef struct ExtractNode:
        lxb_dom_node_t* reference_node
        lxb_tag_id_t tag_id
        size_t depth
        bint space_after
        bint collapse_margins
        bint is_big_block
        bint is_pre
        bint is_end_tag
        shared_ptr[string] text_contents


cdef string _get_collapsed_string(const string& input_str) nogil:
    """
    Collapse newlines and consecutive white space in a string to single spaces.
    Takes into account previously extracted text from ``ctx.text``.
    """
    if input_str.empty():
        return input_str

    cdef string element_text
    element_text.reserve(input_str.size())
    for i in range(input_str.size()):
        if isspace(input_str[i]):
            if element_text.empty() or not isspace(element_text.back()):
                element_text.push_back(b' ')
        else:
            element_text.push_back(input_str[i])

    return element_text


cdef string LIST_BULLET = <const char*>b'\xe2\x80\xa2'


cdef inline void _ensure_text_contents(vector[shared_ptr[ExtractNode]]& extract_nodes) nogil:
    if not deref(extract_nodes.back()).text_contents:
        deref(extract_nodes.back()).text_contents = make_shared[string]()


cdef inline void _ensure_space(string& in_str, char space_char) nogil:
    if in_str.empty() or not isspace(in_str.back()):
        in_str.push_back(space_char)


cdef void _extract_cb(vector[shared_ptr[ExtractNode]]& extract_nodes, ExtractContext& ctx, bint is_end_tag) nogil:
    cdef shared_ptr[ExtractNode] last_node_shared
    cdef ExtractNode* last_node = NULL
    cdef bint is_block = ctx.node.type == LXB_DOM_NODE_TYPE_ELEMENT and is_block_element(ctx.node.local_name)
    if not extract_nodes.empty():
        last_node = extract_nodes.back().get()

    if not last_node or is_block or ctx.depth < last_node.depth or ctx.node.local_name == LXB_TAG_TEXTAREA:
        last_node_shared = make_shared[ExtractNode]()
        extract_nodes.push_back(last_node_shared)
        last_node = extract_nodes.back().get()
        last_node.reference_node = ctx.node
        last_node.depth = ctx.depth
        last_node.is_big_block = ctx.node.local_name in [LXB_TAG_P, LXB_TAG_H1, LXB_TAG_H2, LXB_TAG_H3, LXB_TAG_H4]
        last_node.tag_id = ctx.node.local_name
        last_node.is_pre = not is_end_tag and ctx.node.local_name in [LXB_TAG_PRE, LXB_TAG_TEXTAREA]
        last_node.is_end_tag = is_end_tag

    cdef lxb_dom_character_data_t* char_data = NULL
    cdef string element_text
    cdef string_view element_text_sv

    if ctx.node.type == LXB_DOM_NODE_TYPE_TEXT:
        _ensure_text_contents(extract_nodes)
        node_char_data = <lxb_dom_character_data_t*>ctx.node
        element_text_sv = string_view(<const char*>node_char_data.data.data, node_char_data.data.length)
        if last_node.is_pre and ctx.opts.preserve_formatting:
            deref(last_node.text_contents).append(<string>element_text_sv)
        else:
            element_text = _get_collapsed_string(<string>element_text_sv)
            element_text_sv = <string_view>element_text
            if deref(last_node.text_contents).empty() or isspace(deref(last_node.text_contents).back()):
                while not element_text_sv.empty() and isspace(element_text_sv.front()):
                    element_text_sv.remove_prefix(1)
            if not element_text_sv.empty():
                deref(last_node.text_contents).append(<string>element_text_sv)

    elif ctx.node.type != LXB_DOM_NODE_TYPE_ELEMENT:
        return

    elif ctx.node.local_name in [LXB_TAG_BR, LXB_TAG_HR]:
        _ensure_text_contents(extract_nodes)
        last_node.collapse_margins = False

    elif ctx.opts.links and is_end_tag and ctx.node.local_name == LXB_TAG_A:
        element_text_sv = <string>get_node_attr_sv(ctx.node, b'href')
        if not element_text_sv.empty():
            element_text.append(b' (')
            element_text.append(<string> element_text_sv)
            element_text.append(b') ')
            _ensure_text_contents(extract_nodes)
            deref(last_node.text_contents).append(element_text)

    elif ctx.opts.alt_texts and ctx.node.local_name in [LXB_TAG_IMG, LXB_TAG_AREA]:
        _ensure_text_contents(extract_nodes)
        element_text_sv = get_node_attr_sv(ctx.node, b'alt')
        if not element_text_sv.empty():
            deref(last_node.text_contents).append(<string>element_text_sv)

    elif ctx.opts.form_fields and ctx.node.local_name in [LXB_TAG_TEXTAREA, LXB_TAG_BUTTON]:
        if not is_end_tag:
            element_text.append(b'[ ')
        else:
            element_text.append(b' ]')
        _ensure_text_contents(extract_nodes)
        deref(last_node.text_contents).append(element_text)
        return

    elif ctx.opts.form_fields and ctx.node.local_name == LXB_TAG_INPUT:
        element_text_sv = get_node_attr_sv(ctx.node, b'type')
        if element_text_sv.empty() or element_text_sv not in \
                [b'checkbox', b'color', b'file', b'hidden', b'radio', b'reset']:
            element_text_sv = get_node_attr_sv(ctx.node, b'value')
            if element_text_sv.empty():
                element_text_sv = get_node_attr_sv(ctx.node, b'placeholder')
            if not element_text_sv.empty():
                _ensure_text_contents(extract_nodes)
                element_text.append(b'[ ')
                element_text.append(<string> element_text_sv)
                element_text.append(b' ]')
                _ensure_text_contents(extract_nodes)
                deref(last_node.text_contents).append(element_text)


cdef inline string _indent_newlines(const string& element_text, size_t depth) nogil:
    cdef string indent = string(2 * depth, <char>b' ')
    cdef string tmp_text
    tmp_text.reserve(element_text.size() + 10 * indent.size())
    for i in range(element_text.size()):
        tmp_text.push_back(element_text[i])
        if element_text[i] == b'\n':
            tmp_text.append(indent)

    return tmp_text


cdef string _serialize_extract_nodes(vector[shared_ptr[ExtractNode]]& extract_nodes, const ExtractOpts& opts) nogil:
    cdef size_t i
    cdef string output
    cdef string element_text
    cdef ExtractNode* current_node = NULL
    cdef bint bullet_deferred = False
    cdef size_t list_depth = 0
    cdef vector[size_t] list_numbering
    cdef string list_item_indent = <const char*>b' '

    for i in range(extract_nodes.size()):
        current_node = extract_nodes[i].get()

        if opts.preserve_formatting:
            if current_node.tag_id in [LXB_TAG_UL, LXB_TAG_OL] \
                    or (current_node.tag_id == LXB_TAG_LI and list_depth == 0):
                if current_node.is_end_tag:
                    predec(list_depth)
                    list_numbering.pop_back()
                    bullet_deferred = False
                else:
                    preinc(list_depth)
                    list_numbering.push_back(<size_t>(current_node.tag_id == LXB_TAG_OL))

            if current_node.tag_id == LXB_TAG_LI:
                bullet_deferred = True

            # Add margins
            if current_node.tag_id != LXB_TAG_TEXTAREA:
                if not current_node.collapse_margins or (not output.empty() and output.back() != b'\n'):
                    output.push_back(<char> b'\n')
                if current_node.is_big_block and not bullet_deferred and \
                        output.size() >= 2 and output[output.size() - 2] != b'\n':
                    output.push_back(<char> b'\n')
        elif not output.empty() and output.back() != b' ':
            output.push_back(<char> b' ')

        if current_node.text_contents.get() == NULL:
            continue

        element_text = deref(current_node.text_contents)
        if not current_node.is_pre or current_node.is_end_tag:
            element_text = rstrip_str(element_text)

        if element_text.empty():
            continue

        if list_depth > 0:
            if current_node.is_pre and opts.preserve_formatting:
                element_text = _indent_newlines(element_text, list_depth + <size_t>opts.list_bullets)
            if opts.preserve_formatting:
                list_item_indent = string(2 * list_depth + 2 * <size_t>(
                        not bullet_deferred and opts.list_bullets), <char>b' ')
            if bullet_deferred:
                if opts.list_bullets and list_numbering.back() == 0:
                    list_item_indent += LIST_BULLET + <const char*>b' '
                elif opts.list_bullets:
                    list_item_indent += to_string(list_numbering.back()) + <const char*>b'. '
                    preinc(list_numbering.back())
                bullet_deferred = False
            element_text = list_item_indent + element_text

        if opts.preserve_formatting and current_node.tag_id in [LXB_TAG_TD, LXB_TAG_TH]:
            if not output.empty() and output.back() != b'\n':
                output.append(b'\t\t')

        output.append(element_text)

    return output


cdef inline bint _is_unprintable_pua(lxb_dom_node_t* node) nogil:
    """Whether text node contains only a single unprintable code point from the private use area."""
    if node.first_child and (node.first_child.next or node.first_child.type != LXB_DOM_NODE_TYPE_TEXT):
        # Node has more than one child
        return False
    if not node.first_child and node.type != LXB_DOM_NODE_TYPE_TEXT:
        return False

    cdef string element_text = strip_str(get_node_text(node))
    if element_text.size() > 3:
        return False

    # Pilcrow character (probably an anchor link)
    if element_text == b'\xc2\xb6':
        return True

    # BMP private use area (probably an icon font)
    cdef uint32_t cp = 0
    if element_text.size() == 3:
        memcpy(&cp, element_text.data(), 3 * sizeof(char))
        if 0x8080ee <= cp <= 0xbfa3ef:
            return True

    return False


cdef RE2Options re_opts
re_opts.set_case_sensitive(False)

cdef RE2 article_cls_regex = RE2(rb'(?:^|[\s_-])(?:article|entry|post|story|single[_-]?post|main[_-]content)(?:content|body|text|page)?(?:$|[\s_-])', re_opts)
cdef RE2 nav_cls_regex = RE2(rb'(?:^|\s)(?:(?:l|m|wp|main|site|page|sub|article|global|sticky|main)[_-]*)?(?:nav(?:igation)?|menu(?:[_-]item)?|drop[_-]?down|bread[_-]?crumbs?)|(?:links?[_-]?(?:bar|box|list|container|section|wrapp(?:er))?)(?:$|[\s_-])', re_opts)
cdef RE2 recommended_cls_regex = RE2(rb'(?:^|[\s_-])(?:trends|trending|recommended|featured|popular|editors?[_-]picks|related|read-next|(?:related|more|other)[_-]?(?:links|articles|posts|guides|stories))(?:$|[\s_-])', re_opts)
cdef RE2 landmark_id_regex = RE2(rb'^(?:(?:l|wp|global|page|site|full|sticky)[_-]*)?(?:(?:head|foot)(?:er)?|right)$', re_opts)
cdef RE2 header_cls_regex = RE2(rb'(?:^|\s)(?:l|m|wp|global|page|site|full|sticky)[_-]*header(?:[_-]?wrap(?:per)?|bar)?(?:$|\s)', re_opts)
cdef RE2 footer_cls_regex = RE2(rb'(?:^|[\s_-])(?:global|page|site|copyright)?(?:footer|copyright|cookie|consent|legal|fcontainer)(?:$|[\s_-])', re_opts)
cdef RE2 post_meta_cls_regex = RE2(rb'(?:^|[\s_-])(?:(?:post|entry|article(?:page)?|content|story|section)[_-]*(?:text[_-]*)?(?:footer|teaser|meta(?:[_-]?data)?|subline|sidebar|author(?:name)?|published|timestamp|date|posted[_-]?on|info|labels?|tags?|keywords|category)|by[_-]?line|date[_-]?line|author-date|submitted(?:-by)?)|meta[_-]?data(?:$|[\s_-])', re_opts)
cdef RE2 sidebar_cls_regex = RE2(rb'(?:^|\s)(?:(?:l|wp|right|left|global|sticky)[_-]*)?(?:(?:side|sticky)[_-]?(?:bars?|box)|one-third)(?:$|[\s_-])', re_opts)
cdef RE2 search_cls_regex = RE2(rb'(?:^|[\s_-])search(?:[_-]?(?:bar|facility|box))?(?:$|\s)', re_opts)
cdef RE2 skip_link_cls_regex = RE2(rb'(?:^|\s)(?:link[_-]?)?(?:skip(?:[_-]?(?:to|link))?|scroll[_-]?(?:up|down)|next|prev(?:ious)?|permalink|pagination)(?:$|\s|[_-]?(?:post|article))', re_opts)
cdef RE2 display_cls_regex = RE2(rb'(?:^|\s)(?:is[_-])?(?:display-none|hidden|invisible|collapsed|h-0|nocontent|expandable)(?:-xs|-sm|-lg|-2?xl)?(?:$|\s)', re_opts)
cdef RE2 display_css_regex = RE2(rb'(?:^|;\s*)(?:display\s?:\s?none|visibility\s?:\s?hidden)(?:$|\s?;)', re_opts)
cdef RE2 modal_cls_regex = RE2(rb'(?:^|\s)(?:wp-|p-|-l)?(?:modal|popup|lightbox)(?:[_-]*(?:window|pane|box))?(?:$|[\s_-])', re_opts)
cdef RE2 gallery_cls_regex = RE2(rb'(?:^|[\s_-])(?:gallery|carousel)(?:$|[\s_-])', re_opts)
cdef RE2 signin_cls_regex = RE2(rb'(?:^|[\s_-])(?:(?:log[_-]?in|sign[_-]?(?:in|up)|account)|user[_-](?:info|profile|settings|actions))(?:$|[\s_-])', re_opts)
cdef RE2 ads_cls_regex = RE2(rb'(?:^|\s)(?:(?:google|wide)[_-]?ads?|ad(?:vert|vertise(?:ment|link)?|$|_[a-f0-9]+)|sponsor(?:ed)?|promoted|paid|(?:wide)?banner|donate)(?:$|[\s_-])', re_opts)
cdef RE2 social_cls_regex = RE2(rb'(?:^|\s|__|--|mobile-|desktop-|l-|m-|c-)(?:social(?:media|search)?|share(?:daddy)?|syndication|newsletter|sharing|follow|email|likes?|(?:give[_-]?)?feedback|(?:brand[_-])?engagement|facebook|twitter|subscribe|wa|jp|aptf-follow)(?:[_-]?(?:post|links?|section|icons?|btn|buttons?|target))?(?:$|[\s_-])', re_opts)
cdef RE2 comments_cls_regex = RE2(rb'(?:^|[\s_-])(?:(?:article|user|post)[_-]*)?(?:(?:no[_-]?)?comments?|comment[_-]?list|reply)(?:$|[\s_-])', re_opts)
cdef RE2 logo_cls_regex = RE2(rb'(?:brand(?:ing)?[_-]*)?logo(?:$|\s)', re_opts)
cdef RE2 print_cls_regex = RE2(rb'(?:^|\s)print[_-]', re_opts)


cdef inline bint regex_search_not_empty(const StringPiece& s, const RE2& r) nogil:
    if s.empty():
        return False
    return PartialMatch(s, r())


cdef inline bint _is_link_cluster(lxb_dom_node_t* node, double max_link_ratio, size_t max_length) nogil:
    """
    Check if element contains an excessive number of links compared to the whole content length.
    
    :param node: input node
    :param max_link_ratio: maximum ratio of link chars / all chars
    :param max_length: do not check ratio if content length is larger than this (0 to disable limit)
    :return: true if element is a link cluster
    """
    cdef string element_text = _get_collapsed_string(get_node_text(node))
    if max_length and element_text.size() > max_length:
        return False
    dom_coll = lxb_dom_collection_make(node.owner_document, 20)
    lxb_dom_elements_by_tag_name(<lxb_dom_element_t *> node, dom_coll, <const lxb_char_t *> b'a', 1)
    cdef size_t i
    cdef string link_texts
    link_texts.reserve(element_text.size())
    for i in range(lxb_dom_collection_length(dom_coll)):
        link_texts.append(_get_collapsed_string(get_node_text(lxb_dom_collection_node(dom_coll, i))))
    lxb_dom_collection_destroy(dom_coll, True)
    if not link_texts.empty() and link_texts.size() / <double> element_text.size() > max_link_ratio:
        return True
    return False


cdef stl_set[string] blacklist_aria_roles = [b'alert', b'banner', b'checkbox', b'comment', b'complementary',
                                             b'contentinfo', b'dialog', b'img', b'menu', b'menubar', b'menuitem',
                                             b'navigation', b'presentation', b'radio', b'search', b'searchbox',
                                             b'separator', b'tab', b'toolbar', b'tooltip']


# noinspection DuplicatedCode
cdef inline bint _is_main_content_node(lxb_dom_node_t* node, size_t body_depth, bint allow_comments) nogil:
    """
    Perform a rule-based check whether the given element is a "main-content" element.
    
    :param node: node to check
    :param body_depth: DOM depth of element counted from the document's BODY
    :param allow_comments: treat comment sections as main content
    :return: true if element is a main content element
    """

    if node.type == LXB_DOM_NODE_TYPE_TEXT:
        return not _is_unprintable_pua(node)
    elif node.type != LXB_DOM_NODE_TYPE_ELEMENT:
        return True


    # ------ Section 1: Tag name matching ------

    # Main elements and headings
    if node.local_name in [LXB_TAG_BODY, LXB_TAG_MAIN, LXB_TAG_H1]:
        return True

    # Global footer
    elif node.local_name == LXB_TAG_FOOTER:
        if body_depth < 3 or _is_link_cluster(node, 0.2, 0):
            return False

        # Check if footer is recursive last element node of a direct body child
        pnode = node
        while pnode and pnode.parent and pnode.parent.local_name != LXB_TAG_BODY:
            if pnode.next and pnode.next.type == LXB_DOM_NODE_TYPE_TEXT:
                pnode = pnode.next
            if pnode.next:
                # There is at least one more element node
                return True
            pnode = pnode.parent
        return False

    elif node.local_name == LXB_TAG_UL:
        if body_depth < 4 or _is_link_cluster(node, 0.2, 0):
            return False

    # Teaser articles
    elif node.local_name == LXB_TAG_ARTICLE:
        if body_depth > 2 and _is_link_cluster(node, 0.2, 500):
            return False

    # Navigation, sidebar, other hard-blacklisted elements
    elif node.local_name in [LXB_TAG_NAV, LXB_TAG_ASIDE, LXB_TAG_AUDIO, LXB_TAG_VIDEO, LXB_TAG_TIME]:
        return False


    # ------ Section 2: Rel and ARIA attribute matching ------

    # Hidden elements
    if lxb_dom_element_has_attribute(<lxb_dom_element_t*>node, <const lxb_char_t*>b'hidden', 6):
        return False

    # rel attributes
    cdef string_view rel_attr = get_node_attr_sv(node, b'rel')
    if not rel_attr.empty() and rel_attr in [b'author', b'icon', b'search', b'prev', b'next', b'tag']:
        return False

    # itemprop attributes
    cdef string_view itemprop_attr = get_node_attr_sv(node, b'itemprop')
    if not itemprop_attr.empty() and itemprop_attr in [b'datePublished', b'author', b'url']:
        return False

    # ARIA hidden
    if get_node_attr_sv(node, b'aria-hidden') == b'true':
        return False

    # ARIA expanded
    if get_node_attr_sv(node, b'aria-expanded') == b'false':
        return False


    # ------ Section 3: General class and ID matching ------

    cdef StringPiece cls_attr = get_node_attr_sp(node, b'class')
    cdef StringPiece id_attr = get_node_attr_sp(node, b'id')
    # Only elements with class or id attributes from here on
    if cls_attr.empty() and id_attr.empty():
        if node.local_name == LXB_TAG_DIV:
            return body_depth <= 5 or not _is_link_cluster(node, 0.6, 800)
        return True

    cdef string cls_and_id_attr_str = cls_attr.as_string()
    if not cls_and_id_attr_str.empty():
        cls_and_id_attr_str.push_back(b' ')
    cls_and_id_attr_str.append(id_attr.as_string())
    cdef StringPiece cls_and_id_attr = StringPiece(cls_and_id_attr_str)

    # Hidden elements
    if regex_search_not_empty(cls_attr, display_cls_regex) \
            or regex_search_not_empty(get_node_attr_sp(node, b'style'), display_css_regex):
        return False

    # Skip links
    if node.local_name in [LXB_TAG_A, LXB_TAG_DIV, LXB_TAG_LI] and \
            regex_search_not_empty(cls_and_id_attr, skip_link_cls_regex):
        return False

    if body_depth > 2:
        # Sign-in links
        if regex_search_not_empty(cls_attr, signin_cls_regex):
            return False

        # Post meta
        if regex_search_not_empty(cls_attr, post_meta_cls_regex):
            return False

        # Social media and feedback forms
        if regex_search_not_empty(cls_attr, social_cls_regex):
            return False

    # Logos
    if regex_search_not_empty(cls_and_id_attr, logo_cls_regex):
        return False

    # Ads
    if regex_search_not_empty(cls_and_id_attr, ads_cls_regex) \
            or lxb_dom_element_has_attribute(<lxb_dom_element_t*>node, <const lxb_char_t*>b'data-ad', 7) \
            or lxb_dom_element_has_attribute(<lxb_dom_element_t*>node, <const lxb_char_t*>b'data-advertisement', 18) \
            or lxb_dom_element_has_attribute(<lxb_dom_element_t*>node, <const lxb_char_t*>b'data-text-ad', 12):
        return False


    # ------ Section 4: Class and ID matching of block elements only ------

    if not is_block_element(node.local_name) and node.local_name != LXB_TAG_TD:
        return True

    # ARIA roles
    cdef string_view role_attr = get_node_attr_sv(node, b'role')
    if rel_attr == b'main':
        return True
    if not role_attr.empty() and blacklist_aria_roles.find(<string>role_attr) != blacklist_aria_roles.end():
        return False

    # Whitelist article elements
    if regex_search_not_empty(cls_and_id_attr, article_cls_regex):
        return True

    # Global landmarks by ID
    if regex_search_not_empty(id_attr, landmark_id_regex):
        return False

    # Global header
    if regex_search_not_empty(cls_and_id_attr, header_cls_regex):
        return False

    # Global footer
    if regex_search_not_empty(cls_and_id_attr, footer_cls_regex):
        return False

    # Global navigation
    if regex_search_not_empty(cls_and_id_attr, nav_cls_regex):
        return False

    # Recommended articles
    if regex_search_not_empty(cls_and_id_attr, recommended_cls_regex):
        return False

    # Comments section
    if not allow_comments and node.local_name and regex_search_not_empty(cls_and_id_attr, comments_cls_regex):
        return False

    # Global search bar
    if regex_search_not_empty(cls_and_id_attr, search_cls_regex):
        return False

    # Global sidebar
    if regex_search_not_empty(cls_and_id_attr, sidebar_cls_regex):
        return False

    # Modals
    if regex_search_not_empty(cls_and_id_attr, modal_cls_regex):
        return False

    # Image galleries and carousels
    if regex_search_not_empty(cls_and_id_attr, gallery_cls_regex):
        return False

    # Print content
    if regex_search_not_empty(cls_and_id_attr, print_cls_regex):
        return False

    if body_depth > 2 and node.local_name == LXB_TAG_DIV and _is_link_cluster(node, 0.6, 1500):
        return False

    return True


cdef inline lxb_status_t _collect_selected_nodes_cb(lxb_dom_node_t *node,
                                                    lxb_css_selector_specificity_t *spec, void *ctx) nogil:
    (<stl_set[lxb_dom_node_t*]*>ctx).insert(node)
    return LXB_STATUS_OK


cdef inline lxb_status_t _exists_cb(lxb_dom_node_t *node, lxb_css_selector_specificity_t *spec, void *ctx) nogil:
    (<bint*>ctx)[0] = True
    return LXB_STATUS_STOP


def extract_plain_text(HTMLTree tree, bint preserve_formatting=True, bint main_content=False,
                       bint list_bullets=True,  bint alt_texts=True, bint links=False, bint form_fields=False,
                       bint noscript=False, bint comments=True, skip_elements=None):
    """
    extract_plain_text(tree, preserve_formatting=True, main_content=False, list_bullets=True, alt_texts=False, \
                       links=True, form_fields=False, noscript=False, comments=None, skip_elements=None)

    Perform a simple plain-text extraction from the given DOM node and its children.

    Extracts all visible text (excluding script/style elements, comment nodes etc.)
    and collapses consecutive white space characters. If ``preserve_formatting`` is
    ``True``, line breaks, paragraphs, other block-level elements, list elements, and
    ``<pre>``-formatted text will be preserved.

    Extraction of particular elements and attributes such as links, alt texts, or form fields
    can be be configured individually by setting the corresponding parameter to ``True``.
    Defaults to ``False`` for most elements (i.e., only basic text will be extracted).

    :param tree: HTML DOM tree
    :type tree: DOMNode
    :param preserve_formatting: preserve basic block-level formatting
    :type preserve_formatting: bool
    :param main_content: apply simple heuristics for extracting only "main-content" elements
    :type main_content: bool
    :param list_bullets: insert bullets / numbers for list items
    :type list_bullets: bool
    :param alt_texts: preserve alternative text descriptions
    :type alt_texts: bool
    :param links: extract link target URLs
    :type links: bool
    :param form_fields: extract form fields and their values
    :type form_fields: bool
    :param noscript: extract contents of <noscript> elements
    :param comments: treat comment sections as main content
    :type comments: bool
    :param skip_elements: list of CSS selectors for elements to skip
    :type skip_elements: t.Iterable[str] or None
    :type noscript: bool
    :return: extracted plain text
    :rtype: str
    """
    if not check_node(tree.body):
        return ''

    skip_selectors = {e.encode() for e in skip_elements or []}
    skip_selectors.update({b'script', b'style', b'iframe', b'frame', b'template'})
    if not alt_texts:
        skip_selectors.update({b'object', b'video', b'audio', b'embed' b'img', b'area',
                               b'svg', b'figcaption', b'figure'})
    if not noscript:
        skip_selectors.add(b'noscript')
    if not form_fields:
        skip_selectors.update({b'textarea', b'input', b'button', b'select', b'option', b'label',})

    cdef ExtractContext ctx
    ctx.root_node = <lxb_dom_node_t*>tree.dom_document.body
    ctx.node = ctx.root_node
    ctx.opts = [
        preserve_formatting,
        list_bullets,
        links,
        alt_texts,
        form_fields,
        noscript]

    cdef const lxb_char_t* tag_name = NULL
    cdef size_t tag_name_len
    cdef string tag_name_str
    cdef size_t i
    cdef bint skip = False
    cdef bint is_end_tag = False

    if ctx.node.type == LXB_DOM_NODE_TYPE_DOCUMENT:
        ctx.root_node = next_element_node(ctx.node, ctx.node.first_child)
        ctx.node = ctx.root_node

    cdef lxb_css_parser_t* css_parser = NULL
    cdef lxb_css_selectors_t* css_selectors = NULL
    cdef lxb_selectors_t* selectors = NULL

    cdef lxb_css_selector_list_t* selector_list = NULL
    cdef stl_set[lxb_dom_node_t*] preselected_nodes
    cdef string main_content_selector = b'.article-body, .articleBody, .contentBody, .article-text, .main-content,' \
                                        b'.postcontent, .post-content, .single-post, [role="main"]'

    try:
        init_css_parser(&css_parser)
        init_css_selectors(css_parser, &css_selectors, &selectors)

        if main_content:
            # Opportunistically try to find general main content zone
            selector_list = parse_css_selectors(css_parser, <const lxb_char_t*>main_content_selector.data(),
                                                main_content_selector.size())
            lxb_selectors_find(selectors, ctx.root_node, selector_list,
                               <lxb_selectors_cb_f>_collect_selected_nodes_cb, &preselected_nodes)
            if preselected_nodes.size() == 1:
                # Use result only if there is exactly one match
                ctx.root_node = deref(preselected_nodes.begin())
                ctx.node = ctx.root_node
            preselected_nodes.clear()

        # Select all blacklisted elements and save them in an ordered set
        combined_skip_sel = b','.join(skip_selectors)
        selector_list = parse_css_selectors(css_parser, <const lxb_char_t*>combined_skip_sel, len(combined_skip_sel))
        lxb_selectors_find(selectors, ctx.root_node, selector_list,
                           <lxb_selectors_cb_f>_collect_selected_nodes_cb, &preselected_nodes)
    finally:
        destroy_css_selectors(css_selectors, selectors)
        destroy_css_parser(css_parser)

    cdef size_t base_depth = 0
    cdef lxb_dom_node_t* pnode = ctx.node
    while pnode.local_name != LXB_TAG_BODY and pnode.parent:
        preinc(base_depth)
        pnode = pnode.parent

    cdef vector[shared_ptr[ExtractNode]] extract_nodes
    extract_nodes.reserve(150)
    with nogil:
        while ctx.node:
            # Skip everything except element and text nodes
            if ctx.node.type != LXB_DOM_NODE_TYPE_ELEMENT and ctx.node.type != LXB_DOM_NODE_TYPE_TEXT:
                is_end_tag = True
                ctx.node = next_node(ctx.root_node, ctx.node, &ctx.depth, &is_end_tag)
                continue

            # Skip blacklisted or non-main-content nodes
            if preselected_nodes.find(ctx.node) != preselected_nodes.end() or \
                    (main_content and not _is_main_content_node(ctx.node, ctx.depth + base_depth, comments)):
                is_end_tag = True
                ctx.node = next_node(ctx.root_node, ctx.node, &ctx.depth, &is_end_tag)
                continue

            _extract_cb(extract_nodes, ctx, is_end_tag)

            ctx.node = next_node(ctx.root_node, ctx.node, &ctx.depth, &is_end_tag)

    return bytes_to_str(_serialize_extract_nodes(extract_nodes, ctx.opts)).rstrip()
